@@ -1,22 +1,21 @@
-# $Id: ObjectDriver.pm 361 2007-05-02 04:45:44Z btrott $
+# $Id$
 
 package Data::ObjectDriver;
 use strict;
 use warnings;
+use 5.006_001;
 use Class::Accessor::Fast;
 
 use base qw( Class::Accessor::Fast );
+use Data::ObjectDriver::Iterator;
 
-use Data::ObjectDriver::Profiler;
+__PACKAGE__->mk_accessors(qw( pk_generator txn_active ));
 
-__PACKAGE__->mk_accessors(qw( pk_generator ));
-
-our $VERSION = '0.04';
+our $VERSION = '0.07';
 our $DEBUG = $ENV{DOD_DEBUG} || 0;
 our $PROFILE = $ENV{DOD_PROFILE} || 0;
-our $PROFILER = Data::ObjectDriver::Profiler->new;
-
-use Data::Dumper ();
+our $PROFILER;
+our $LOGGER;
 
 sub new {
     my $class = shift;
@@ -25,10 +24,22 @@ sub new {
     $driver;
 }
 
+sub logger {
+    my $class = shift;
+    if ( @_ ) {
+        return $LOGGER = shift;
+    } else {
+        return $LOGGER ||= sub {
+            print STDERR @_;
+        };
+    }
+}
+
 sub init {
     my $driver = shift;
     my %param = @_;
     $driver->pk_generator($param{pk_generator});
+    $driver->txn_active(0);
     $driver;
 }
 
@@ -40,37 +51,69 @@ sub start_query {
     my($sql, $bind) = @_;
 
     $driver->debug($sql, $bind) if $DEBUG;
-    $PROFILER->record_query($driver, $sql) if $PROFILE;
+    $driver->profiler($sql) if $PROFILE;
 
     return;
 }
 
 sub end_query { }
 
+sub begin_work {
+    my $driver = shift;
+    $driver->txn_active(1);
+    $driver->debug(sprintf("%14s", "BEGIN_WORK") . ": driver=$driver");
+}
+
+sub commit {
+    my $driver = shift;
+    _end_txn($driver, 'commit');
+}
+
+sub rollback {
+    my $driver = shift;
+    _end_txn($driver, 'rollback');
+}
+
+sub _end_txn {
+    my $driver = shift;
+    my $method = shift;
+    $driver->txn_active(0);
+    $driver->debug(sprintf("%14s", uc($method)) . ": driver=$driver");
+}
+
 sub debug {
     my $driver = shift;
     return unless $DEBUG;
 
-    my $class = ref $driver;
+    my $class = ref $driver || $driver;
     my @caller;
     my $i = 0;
     while (1) {
         @caller = caller($i++);
-        last if $caller[0] !~ /^(Data::ObjectDriver|$driver)/;
+        last if $caller[0] !~ /^(Data::ObjectDriver|$class)/;
     }
 
     my $where = " in file $caller[1] line $caller[2]\n";
 
     if (@_ == 1 && !ref($_[0])) {
-        print STDERR @_, $where;
+        $driver->logger->( @_, $where );
     } else {
+        require Data::Dumper;
         local $Data::Dumper::Indent = 1;
-        print STDERR Data::Dumper::Dumper(@_), $where;
+        $driver->logger->( Data::Dumper::Dumper(@_), $where );
     }
 }
 
 sub profiler {
-    return $PROFILER;
+    my $driver = shift;
+    my ($sql) = @_;
+    $PROFILER ||= eval {
+        require Data::ObjectDriver::Profiler;
+        Data::ObjectDriver::Profiler->new;
+    };
+    return $PROFILE = 0 if $@ || !$PROFILER;
+    return $PROFILER unless @_;
+    $PROFILER->record_query($driver, $sql);
 }
 
 sub list_or_iterator {
@@ -82,11 +125,13 @@ sub list_or_iterator {
     if (wantarray) {
         return @{$objs};
     } else {
-        return sub { shift @{$objs} };
+        my $iter = sub { shift @{$objs} };
+        return Data::ObjectDriver::Iterator->new($iter);
     }
 }
 
 sub cache_object { }
+sub uncache_object { }
 
 1;
 __END__
@@ -157,11 +202,6 @@ caching and partitioning layers, allowing you to spread data across multiple
 physical databases, without your application code needing to know where the
 data is stored.
 
-It's currently considered ALPHA code. The API is largely fixed, but may seen
-some small changes in the future. For what it's worth, the likeliest area
-for changes are in the syntax for the I<search> method, and would most
-likely not break much in the way of backwards compatibility.
-
 =head1 METHODOLOGY
 
 I<Data::ObjectDriver> provides you with a framework for building
@@ -209,7 +249,7 @@ to the driver listed in the I<fallback> setting: the partitioning layer.
 =item 2.
 
 The partitioning layer does not know how to look up objects by itself--all
-it knows how to do is to give back a driver that I<does> know how to loko
+it knows how to do is to give back a driver that I<does> know how to look
 up objects in a backend datastore.
 
 In our example above, imagine that we're partitioning our ingredient data
@@ -325,6 +365,14 @@ a subroutine that acts as an iterator object, like so:
         ...
     }
 
+C<$iter> is blessed in L<Data::ObjectDriver::Iterator> package, so the above
+could also be written:
+
+    my $iter = Ingredient->search({ recipe_id => 5 });
+    while (my $ingredient = $iter->next()) {
+        ...
+    }
+
 The keys in I<%terms> should be column names for the database table
 modeled by I<Class> (and the values should be the desired values for those
 columns).
@@ -369,8 +417,39 @@ Optional; the default is to fetch the values of all of the columns.
 If set to a true value, the I<SELECT> statement generated will include a
 I<FOR UPDATE> clause.
 
+=item * comment
+
+A sql comment to watermark the SQL query.
+
+=item * window_size
+
+Used when requesting an iterator for the search method and selecting
+a large result set or a result set of unknown size. In such a case,
+no LIMIT clause is assigned, which can load all available objects into
+memory. Specifying C<window_size> will load objects in manageable chunks.
+This will also cause any caching driver to be bypassed for issuing
+the search itself. Objects are still placed into the cache upon load.
+
+This attribute is ignored when the search method is invoked in an array
+context, or if a C<limit> attribute is also specified that is smaller than
+the C<window_size>.
+
 =back
 
+=head2 Class->search(\@terms [, \%options ])
+
+This is an alternative calling signature for the search method documented
+above. When providing an array of terms, it allows for constructing complex
+expressions that mix 'and' and 'or' clauses. For example:
+
+    my $iter = Ingredient->search([ { recipe_id => 5 },
+        -or => { calories => { value => 300, op => '<' } } ]);
+    while (my $ingredient = $iter->()) {
+        ...
+    }
+
+Supported logic operators are: '-and', '-or', '-and_not', '-or_not'.
+
 =head2 Class->add_trigger($trigger, \&callback)
 
 Adds a trigger to all objects of class I<Class>, such that when the event
@@ -530,6 +609,89 @@ I<Data::ObjectDriver::Profiler> instance:
 Then see the documentation for I<Data::ObjectDriver::Profiler> to see the
 methods on that class.
 
+
+=head1 TRANSACTIONS
+
+
+Transactions are supported by Data::ObjectDriver's default drivers. So each
+Driver is capable to deal with transactional state independently. Additionally
+<Data::ObjectDriver::BaseObject> class know how to turn transactions switch on
+for all objects.
+
+In the case of a global transaction all drivers used during this time are put
+in a transactional state until the end of the transaction.
+
+=head2 Example
+
+    ## start a transaction
+    Data::ObjectDriver::BaseObject->begin_work;
+
+    $recipe = Recipe->new;
+    $recipe->title('lasagnes');
+    $recipe->save;
+
+    my $ingredient = Ingredient->new;
+    $ingredient->recipe_id($recipe->recipe_id);
+    $ingredient->name("more layers");
+    $ingredient->insert;
+    $ingredient->remove;
+
+    if ($you_are_sure) {
+        Data::ObjectDriver::BaseObject->commit;
+    }
+    else {
+        ## erase all trace of the above
+        Data::ObjectDriver::BaseObject->rollback;
+    }
+
+=head2 Driver implementation
+
+Drivers have to implement the following methods:
+
+=over 4
+
+=item * begin_work to initialize a transaction
+
+=item * rollback
+
+=item * commmit
+
+=back
+
+=head2 Nested transactions
+
+Are not supported and will result in warnings and the inner transactions
+to be ignored. Be sure to B<end> each transaction and not to let et long
+running transaction open (i.e you should execute a rollback or commit for
+each open begin_work).
+
+=head2 Transactions and DBI
+
+In order to make transactions work properly you have to make sure that
+the C<$dbh> for each DBI drivers are shared among drivers using the same
+database (basically dsn).
+
+One way of doing that is to define a get_dbh() subref in each DBI driver
+to return the same dbh if the dsn and attributes of the connection are
+identical.
+
+The other way is to use the new configuration flag on the DBI driver that
+has been added specifically for this purpose: C<reuse_dbh>.
+
+    ## example coming from the test suite
+    __PACKAGE__->install_properties({
+        columns => [ 'recipe_id', 'partition_id', 'title' ],
+        datasource => 'recipes',
+        primary_key => 'recipe_id',
+        driver => Data::ObjectDriver::Driver::Cache::Cache->new(
+            cache => Cache::Memory->new,
+            fallback => Data::ObjectDriver::Driver::DBI->new(
+                dsn      => 'dbi:SQLite:dbname=global.db',
+                reuse_dbh => 1,  ## be sure that the corresponding dbh is shared
+            ),
+        ),
+    });
+
 =head1 EXAMPLES
 
 =head2 A Partitioned, Caching Driver
@@ -587,14 +749,44 @@ methods on that class.
 
     1;
 
+=head1 SUPPORTED DATABASES
+
+I<Data::ObjectDriver> is very modular and it's not very diffucult to add new drivers.
+
+=over 4
+
+=item * MySQL is well supported and has been heavily tested.
+
+=item * PostgreSQL has been been used in production and should just work, too.
+
+=item * SQLite is supported, but YMMV depending on the version. This is the
+backend used for the test suite.
+
+=item * Oracle support has been added in 0.06
+
+=back
+
 =head1 LICENSE
 
 I<Data::ObjectDriver> is free software; you may redistribute it and/or modify
 it under the same terms as Perl itself.
 
+=head1 MAILING LIST, CODE & MORE INFORMATION
+
+I<Data::ObjectDriver> developers can be reached via the following group:
+L<http://groups.google.com/group/data-objectdriver>
+
+Bugs should be reported using the CPAN RT system, patches are encouraged when
+reporting bugs.
+
+L<http://code.sixapart.com/>
+
+Alternatively you can fork our git repositories. See the full list at:
+http://github.com/sixapart
+
 =head1 AUTHOR & COPYRIGHT
 
-Except where otherwise noted, I<Data::ObjectDriver> is Copyright 2005-2006
+Except where otherwise noted, I<Data::ObjectDriver> is Copyright 2005-2010
 Six Apart, cpan@sixapart.com. All rights reserved.
 
 =cut
